# -*- coding: utf-8 -*-

"""
    Module to mange Google Compute Engine.
"""


import logging
from copy import deepcopy
from time import sleep
from .service import get_compute


class Gce(object):

    def __init__(self, oauth_file="", service_account_file="", project="",
                 zone=""):
        self.oauth_file = oauth_file
        self.service_account_file = service_account_file
        self.project = project
        self.zone = zone
        self.instances = {}
        self.compute_service = None
        self.n_wait = 100
        self.wait_time = 10
        self.logger = logging.getLogger(__name__)

    def compute(self):
        if self.compute_service is None:
            self.compute_service = get_compute(
                service_account_file=self.service_account_file,
                oauth_file=self.oauth_file,
            )
        return self.compute_service

    def get_zones(self, **zone_filter):
        zones = []
        for zone in self.compute().zones().list(
                project=self.project).execute()["items"]:
            is_ok = 1
            for k, v in zone_filter.items():
                if zone[k] != v:
                    is_ok = 0
                    continue
            if is_ok == 1:
                zones.append(zone["name"])
        return zones

    def get_instance(self, instance, update=True):
        if update:
            try:
                info = self.compute().instances().get(
                    project=self.project,
                    zone=self.zone,
                    instance=instance).execute()
            except Exception:
                info = None
        else:
            if instance in self.instances:
                info = self.instances[instance]
            else:
                info = None
        return info

    def update_instances(self):
        self.instances = {x["name"]: x for x in
                          self.compute().instances().list(
                              project=self.project,
                              zone=self.zone).execute()["items"]}

    def get_instances(self, update=True, instance_filter={}):
        if update:
            self.update_instances()
        instances = {}
        for instance, info in self.instances.items():
            is_ok = 1
            for k, v in instance_filter.items():
                if info[k] != v:
                    is_ok = 0
                    continue
            if is_ok == 1:
                instances[instance] = info
        return instances

    def check_instance(self, instance, status="RUNNING", n_wait=-1,
                       wait_time=-1, update=True):
        if n_wait == -1:
            n_wait = self.n_wait
        if wait_time == -1:
            wait_time = self.wait_time
        if n_wait <= 0:
            return True
        wait = n_wait
        info = self.get_instance(instance, update)
        while True:
            if status == "DELETED":
                if info is None:
                    return True
            elif status == "INSERTED":
                if info is not None:
                    return True
            else:
                if info is None:
                    continue
                if info["status"] == status:
                    return True
            wait = wait - 1
            if wait <= 0:
                break
            sleep(wait_time)
            info = self.get_instance(instance)
        return False

    def start_instance(self, instance, n_wait=-1, wait_time=-1, update=True):
        self.logger.debug("Starting %s" % (instance))
        if not self.check_instance(instance, "TERMINATED", 1, 1, update):
            self.logger.warning("%s is not TERMINATED status (status=%s)" %
                                (instance, self.instances[instance]["status"]))
            return False
        self.compute().instances().start(project=self.project,
                                         zone=self.zone,
                                         instance=instance).execute()
        return self.check_instance(instance, "RUNNING", n_wait, wait_time)

    def stop_instance(self, instance, n_wait=-1, wait_time=-1, update=True):
        self.logger.debug("Stopping %s" % (instance))
        if not self.check_instance(instance, "RUNNING", 1, 1, update):
            self.logger.warning("%s is not RUNNING status (status=%s)" %
                                (instance, self.instances[instance]["status"]))
            return False
        self.compute().instances().stop(project=self.project,
                                        zone=self.zone,
                                        instance=instance).execute()
        return self.check_instance(instance, "TERMINATED", n_wait, wait_time)

    def get_source_disk_image(self, family, project=""):
        if project == "":
            project = self.project
        image_response = self.compute().images().getFromFamily(
            project=project, family=family).execute()
        return image_response['selfLink']

    def insert_instance(self, instance, n_wait=-1, wait_time=-1, option={},
                        update=True):
        self.logger.debug("Inserting %s" % (instance))
        if self.check_instance(instance, "INSERTED", 1, 1, update):
            self.logger.warning("%s already exists" % instance)
            return False
        opt = deepcopy(option)
        if "name" not in opt:
            opt["name"] = instance
        if not opt["machineType"].startswith("zones"):
            opt["machineType"] = "zones/%s/machineTypes/%s" % (
                self.zone, opt["machineType"])
        if "disks" not in opt:
            opt["disks"] = [{}]
            opt["disks"][0] = {
                "boot": True,
                "autoDelete": True,
            }
        if "initializeParams" not in opt["disks"][0]:
            opt["disks"][0]["initializeParams"] = {}
        if "family" in opt:
            if "sourceImage" not in opt["disks"][0]["initializeParams"]:
                if "project" in opt:
                    project = opt["project"]
                else:
                    project = self.project
                source_disk_image = self.get_source_disk_image(opt["family"],
                                                               project)
                opt["disks"][0]["initializeParams"]["sourceImage"] =\
                    source_disk_image
            del opt["family"]
        if "networkInterfaces" not in opt:
            opt["networkInterfaces"] = [{
                "network": "global/networks/default",
                "accessConfigs": [
                    {"type": "ONE_TO_ONE_NAT", "name": "External NAT"}
                ]
            }]
        self.compute().instances().insert(project=self.project,
                                          zone=self.zone,
                                          body=opt).execute()
        if opt["disks"][0]["boot"]:
            return self.check_instance(instance, "RUNNING", n_wait, wait_time)
        else:
            return self.check_instance(instance, "INSERTED", n_wait, wait_time)

    def create_instance(self, instance, n_wait=-1, wait_time=-1, option={},
                        update=True):
        return self.insert_instance(instance, n_wait, wait_time, option,
                                    update)

    def delete_instance(self, instance, n_wait=-1, wait_time=-1, update=True):
        self.logger.debug("Deleting %s" % (instance))
        if self.check_instance(instance, "DELETED", 1, 1, update):
            self.logger.warning("%s does not exist)" % instance)
            return False
        self.compute().instances().delete(
            project=self.project,
            zone=self.zone,
            instance=instance,
        ).execute()
        return self.check_instance(instance, "DELETED", n_wait, wait_time)



# -*- coding: utf-8 -*-

"""
    Core module to provides gcpm functions.
"""


import os
import logging
import copy
import json
import ruamel.yaml
from time import sleep
from pprint import pformat
from googleapiclient.errors import HttpError
from .__init__ import __version__
from .__init__ import __program__
from .utils import expand, make_startup_script, make_shutdown_script
from .files import make_file, make_service, make_logrotate, rm_service,\
    rm_logrotate
from .condor import Condor
from .gce import Gce
from .gcs import Gcs


class Gcpm(object):
    """HTCondor pool manager for Google Cloud Platform."""

    def __init__(self, config="", service=False, test=False):
        self.first_update_config = 0
        self.logger = None
        self.is_service = service
        if config == "":
            if self.is_service:
                self.config = "/etc/gcpm.yml"
            else:
                self.config = "~/.config/gcpm/gcpm.yml"
        else:
            self.config = config
        self.config = expand(self.config)

        self.data = {
            "config_dir": "",
            "oauth_file": "",
            "wn_list": "",
            "service_account_file": "",
            "project": "",
            "zone": "",
            "max_cores": 0,
            "machines": [],
            "static_wns": [],
            "required_machines": [],
            "primary_accounts": [],
            "prefix": "gcp-wn",
            "instance_max_num": 9999,
            "preemptible": 0,
            "off_timer": 0,
            "network_tag": [],
            "reuse": 0,
            "interval": 10,
            "head_info": "gcp",
            "head": "",
            "port": 9618,
            "domain": "",
            "admin": "",
            "owner": "",
            "wait_cmd": 0,
            "bucket": "",
            "storageClass": "REGIONAL",
            "location": "",
            "log_file": None,
            "log_level": logging.INFO,
        }
        if self.is_service:
            self.data["log_file"] = "/var/log/gcpm.log"

        self.scripts = {"wn": {"startup": {}, "shutdown": {}},
                        "test_wn": {"startup": {}, "shutdown": {}}}
        self.prefix_core = {}
        self.test_prefix_core = {}
        self.services = {}
        self.gce = None
        self.gcs = None
        self.n_wait = 0
        self.create_option = ""
        self.instances_gce = {}
        self.wns = {}
        self.prev_wns = {}
        self.wn_list = ""
        self.condor_wns = []
        self.wn_starting = []
        self.wn_deleting = []
        self.full_idle_jobs = {}
        self.test_idle_jobs = {}
        self.wn_status = {}
        self.total_core_use = []

        self.test = test
        self.condor = Condor(test=self.test)

        self.update_config()

    @staticmethod
    def help():
        print("""
Usage: gcpm [--config=<config>] [--test=<test>] [--oneshot=<oneshot>] <command>

    commands:
        run        : Run user process.
        service    : Run service process.
        set_pool_password <pool_password> : \
Upload pool_password file to Google Cloud Storage.
        install    : Install service (systemd) \
and logrotate configuration file.
        uninstall  : Uninstall service (systemd) \
and logrotate configuration file.
        show_config: Show configurations.
        version    : Show version.
        help       : Show this help.

    options:
        config  : Configuration file for gcpm.
                  Default: ~/.config/gcpm/gcpm.yml for user process.
                           /etc/gcpm.yml for service process.
        oneshot : Set True to run only one loop.
        test    : Set True to test on machines \
which does not have HTCondor service.
""")

    @staticmethod
    def version():
        print("%s: %s" % (__program__, __version__))

    def check_config(self):
        return True

    def read_config(self):
        yaml = ruamel.yaml.YAML()
        if not os.path.isfile(self.config):
            print("GCPM setting file: %s does not exist" % self.config)
        else:
            with open(self.config) as stream:
                data = yaml.load(stream)
            for k, v in data.items():
                self.data[k] = v

        if self.data["config_dir"] == "":
            if self.is_service:
                config_dir = "/var/cache/gcpm"
            else:
                config_dir = "~/.config/gcpm"
            self.data["config_dir"] = expand(config_dir)
        if self.data["oauth_file"] == "":
            self.data["oauth_file"] = self.data["config_dir"] + "/oauth"
        if self.data["wn_list"] == "":
            self.data["wn_list"] = self.data["config_dir"] + "/wn_list.json"

        self.prefix_core = {}
        self.test_prefix_core = {}
        for machine in self.data["machines"]:
            self.prefix_core[machine["core"]] = \
                "%s-%dcore" % (self.data["prefix"], machine["core"])
            self.test_prefix_core[machine["core"]] = \
                "%s-test-%dcore" % (self.data["prefix"], machine["core"])
        if self.data["location"] == "":
            if self.data["storageClass"] == "MULTI_REGIONAL":
                self.data["location"] = self.data["zone"].split("-")[0]
            else:
                self.data["location"] = "-".join(
                    self.data["zone"].split("-")[0:2])
        if self.data["bucket"] == "":
            self.data["bucket"] = self.data["project"] + "_" + "gcpm_bucket"
        if self.data["bucket"].startswith("gs://"):
            self.data["bucket"] = self.data["bucket"].replace("gs://", "")
        if self.data["head"] == "":
            if self.data["head_info"] == "hostname":
                self.data["head"] = os.uname()[1]
            elif self.data["head_info"] == "ip":
                import socket
                self.data["head"] = socket.gethostbyname(socket.gethostname())
            elif self.data["head_info"] == "gcp":
                self.data["head"] = os.uname()[1]
            else:
                raise ValueError(
                    "Both %s and %s are empty" % ("head", "head_info"))
        if self.data["domain"] == "":
            self.data["domain"] = ".".join(os.uname()[1].split(".")[1:])
        if self.data["wait_cmd"] == 1:
            self.n_wait = 100

        if self.data["log_level"].isdigit():
            self.data["log_level"] = int(self.data["log_level"])
        if type(self.data["log_level"]) is int:
            self.data["log_level"] = logging.getLevelName(
                self.data["log_level"])
        self.data["log_level"] = self.data["log_level"].upper()

    def set_logger(self):
        if self.logger is None:
            log_options = {
                "format": '%(asctime)s %(message)s',
                "datefmt": '%Y-%m-%d %H:%M:%S',
            }
            if self.data["log_file"] is not None:
                log_options["filename"] = self.data["log_file"]
            log_options["level"] = self.data["log_level"]

            logging.basicConfig(**log_options)
            self.logger = logging.getLogger(__name__)

        self.logger.setLevel(self.data["log_level"])
        if self.data["log_level"] not in ["NOTSET", "DEBUG"]:
            logging.getLogger("googleapiclient.discovery").setLevel("WARNING")
        else:
            logging.getLogger("googleapiclient.discovery").setLevel(
                self.data["log_level"])

    def show_config(self):
        if self.logger is not None:
            self.logger.info(
                "Configurations have been updated:\n" + pformat(self.data))

    def make_scripts(self):
        for wn_type in self.scripts:
            for machine in self.data["machines"]:
                self.scripts[wn_type]["startup"][machine["core"]] \
                    = make_startup_script(
                        core=machine["core"],
                        mem=machine["mem"],
                        disk=machine["disk"],
                        image=machine["image"],
                        preemptible=self.data["preemptible"],
                        admin=self.data["admin"],
                        head=self.data["head"],
                        port=self.data["port"],
                        domain=self.data["domain"],
                        owner=self.data["owner"],
                        bucket=self.data["bucket"],
                        off_timer=self.data["off_timer"],
                        wn_type=wn_type
                    )
                self.scripts[wn_type]["shutdown"][machine["core"]] \
                    = make_shutdown_script()

    def after_update_config(self):
        self.set_logger()
        self.show_config()
        self.make_scripts()

    def update_config(self):
        orig_data = copy.deepcopy(self.data)
        self.read_config()
        self.check_config()
        if self.first_update_config != 0 and orig_data == self.data:
            return
        self.first_update_config = 1
        self.after_update_config()

    def install(self):
        make_service()
        make_logrotate(mkdir=False)

    def uninstall(self):
        rm_service()
        rm_logrotate()

    def get_gce(self):
        if self.gce is None:
            self.gce = Gce(
                oauth_file=self.data["oauth_file"],
                service_account_file=self.data["service_account_file"],
                project=self.data["project"],
                zone=self.data["zone"],
            )
        return self.gce

    def get_gcs(self):
        if self.gcs is None:
            self.gcs = Gcs(
                oauth_file=self.data["oauth_file"],
                service_account_file=self.data["service_account_file"],
                project=self.data["project"],
                storageClass=self.data["storageClass"],
                location=self.data["location"],
                bucket=self.data["bucket"],
            )
        return self.gcs

    def check_condor_status(self):
        if self.condor.condor_status()[0] != 0:
            raise RuntimeError("HTCondor is not running!")

    def check_required(self):
        for machine in self.data["required_machines"]:
            if machine["name"] in self.instances_gce:
                status = self.instances_gce[machine["name"]]["status"]
                if status == "RUNNING":
                    continue
                elif status == "TERMINATED":
                    if not self.get_gce().start_instance(machine["name"],
                                                         n_wait=100,
                                                         update=False):
                        raise RuntimeError(
                            "Failed to start required machine: %s"
                            % machine["name"])
                else:
                    raise RuntimeError(
                        "Required machine %s is unknown stat: %s"
                        % (machine["name"], status))
            else:
                if not self.new_instance(machine["name"], machine, n_wait=100,
                                         update=True, wn_type=None):
                    raise RuntimeError(
                        "Failed to create required machine: %s"
                        % machine["name"])

    def get_instances_gce(self, update=True):
        if update:
            self.instances_gce = self.get_gce().get_instances(update=True)
        return self.instances_gce

    def get_instances_wns(self, update=True):
        self.get_instances_gce(update)
        instances = {}
        for instance, info in self.instances_gce.items():
            is_use = 0
            for core, prefix in list(self.prefix_core.items()) \
                    + list(self.test_prefix_core.items()):
                if instance.startswith(prefix):
                    is_use = 1
                    break
            for static in self.data["static_wns"]:
                if instance == static:
                    is_use = 1
            if is_use:
                instances[instance] = info
        return instances

    def get_instances_running(self, update=True):
        return {x: y for x, y in self.get_instances_wns(update=update).items()
                if y["status"] == "RUNNING"}

    def get_instances_non_terminated(self, update=True):
        return {x: y for x, y in self.get_instances_wns(update=update).items()
                if y["status"] != "TERMINATED"}

    def get_instances_terminated(self, update=True):
        return {x: y for x, y in self.get_instances_wns(update=update).items()
                if y["status"] == "TERMINATED"}

    def get_gce_ip(self, instance, update=True):
        if instance not in self.get_instances_running(update=update):
            return instance
        info = self.get_instances_running(update=False)[instance]
        if self.data["head_info"] == "gcp":
            ip = info["networkInterfaces"][0]["networkIP"]
        else:
            ip = info["networkInterfaces"][0]["accessConfigs"][0]["natIP"]
        return ip

    def add_gce_wns(self, update=True):
        for instance, info in \
                self.get_instances_running(update=update).items():
            self.wns[instance] = self.get_gce_ip(instance, update=False)

    def add_remaining_wns(self):
        # Check instance which is not running, but in condor_status
        # (should be in the list until it is removed from the status)
        for wn in self.condor_wns:
            if wn not in self.wns:
                if wn in self.prev_wns:
                    self.wns[wn] = self.prev_wns[wn]
                else:
                    self.logger.warning(
                        "%s is listed in the condor status, "
                        "but no information can be taken from gce" % (wn))

    def make_wn_list(self):
        self.wns = {}
        self.prev_wns = {}
        if os.path.isfile(self.data["wn_list"]):
            with open(self.data["wn_list"]) as f:
                self.prev_wns = json.load(f)

        for s in self.data["static_wns"]:
            self.wns[s] = self.get_gce_ip(s, update=False)

        self.add_gce_wns(update=False)
        self.add_remaining_wns()

        self.wn_list = ""
        for name, ip in self.wns.items():
            self.wn_list += \
                " condor@$(UID_DOMAIN)/%s condor_pool@$(UID_DOMAIN)/%s" \
                % (ip, ip)
        make_file(filename=self.data["wn_list"], content=json.dumps(self.wns),
                  mkdir=True)

    def update_condor_collector(self):
        self.make_wn_list()
        self.condor.condor_config_val(["-collector", "-set",
                                       "WNS = %s" % self.wn_list])
        self.condor.condor_reconfig(["-collector"])

    def stop_instance(self, instance):
        self.wn_deleting.append(instance)
        try:
            self.get_gce().stop_instance(instance,
                                         n_wait=self.n_wait,
                                         update=False)
        except HttpError as e:
            self.wn_deleting.remove(instance)
            self.logger.warning(e)

    def delete_instance(self, instance):
        self.wn_deleting.append(instance)
        try:
            self.get_gce().delete_instance(instance,
                                           n_wait=self.n_wait,
                                           update=False)
        except HttpError as e:
            self.wn_deleting.remove(instance)
            self.logger.warning(e)

    def clean_wns(self):
        self.logger.debug("clean_wns")
        for wn in self.wn_starting:
            if wn in self.condor_wns:
                self.wn_starting.remove(wn)

        exist_list = self.data["static_wns"] + self.condor_wns \
            + self.wn_starting + self.wn_deleting
        instances = []

        # Delete condor_off instances
        for instance, info in self.get_instances_wns(update=False).items():
            if self.data["reuse"] and info["status"] == "TERMINATED":
                continue
            instances.append(instance)
            if instance in exist_list:
                continue
            if info["status"] not in ["RUNNING", "TERMINATED"]:
                continue
            if self.data["reuse"]:
                self.stop_instance(instance)
            else:
                self.delete_instance(instance)

        for wn in self.wn_deleting:
            if wn not in instances:
                self.wn_deleting.remove(wn)

    def check_terminated(self):
        if self.data["reuse"] == 1:
            return
        for instance, info in self.get_instances_terminated(
                update=False).items():
            if instance in self.wn_starting + self.wn_deleting:
                continue
            self.delete_instance(instance)

    def update_total_core_use(self):
        working = list(self.get_instances_non_terminated(update=False)) \
            + self.wn_starting + self.wn_deleting

        self.total_core_use = 0
        for wn in working:
            for core, prefix in list(self.prefix_core.items()) \
                    + list(self.test_prefix_core.items()):
                if wn.startswith(prefix):
                    self.total_core_use += core
                    break

    def get_full_wns(self):
        return list(self.instances_gce) + self.wn_starting \
            + self.wn_deleting + self.condor_wns

    def check_wns(self):
        self.check_terminated()
        self.update_total_core_use()

    def update_wns(self):
        self.get_instances_wns(update=False)
        self.condor_wns = self.condor.condor_wn()
        self.update_condor_collector()
        self.clean_wns()
        self.check_wns()

    def check_for_core(self, machine):
        core = machine["core"]
        n_idle_jobs = self.full_idle_jobs[core] \
            if core in self.full_idle_jobs else 0
        machines = {x: y for x, y in self.wn_status.items()
                    if x.startswith(self.prefix_core[core])}
        n_machines = len(machines)
        unclaimed = [x for x, y in machines.items() if y == "Unclaimed"]
        n_unclaimed = len(unclaimed) - n_idle_jobs - machine["idle"]

        for wn in self.wn_starting:
            if wn.startswith(self.prefix_core[core]):
                n_machines += 1
                n_unclaimed += 1

        if n_machines >= machine["max"]:
            return False

        if n_unclaimed >= 0:
            return False

        if self.data["max_cores"] > 0:
            if self.total_core_use + core > self.data["max_cores"]:
                return False

        return True

    def start_terminated(self, core):
        if self.data["reuse"] != 1:
            return False

        for instance in self.get_instances_non_terminated(update=False):
            if instance.startswith(self.prefix_core[core]):
                self.wn_starting.append(instance)
                try:
                    self.get_gce().start_instance(instance, n_wait=self.n_wait,
                                                  update=False)
                except HttpError as e:
                    self.wn_starting.remove(instance)
                    self.logger.warning(e)
                    return False
                return True
        return True

    def new_instance(self, instance_name, machine, n_wait=0,
                     update=False, wn_type=None):
        # memory must be N x 256 (MB)
        q, mod = divmod(machine["mem"], 256)
        memory = q * 256
        if mod != 0:
            memory += 256

        option = {
            "name": instance_name,
            "machineType": "custom-%d-%d" % (machine["core"], memory),
            "disks": [
                {
                    "type": "PERSISTENT",
                    "boot": True,
                    "autoDelete": True,
                    "initializeParams": {
                        "diskSizeGb": machine["disk"],
                        "sourceImage": "global/images/" + machine["image"],
                    }
                }
            ],
            "serviceAccounts": [{
                "email": "default",
                "scopes": [
                    "https://www.googleapis.com/auth/devstorage.read_only",
                    "https://www.googleapis.com/auth/logging.write",
                    "https://www.googleapis.com/auth/monitoring.write",
                    "https://www.googleapis.com/auth/trace.append",
                ]
            }],
        }
        if wn_type is not None:
            option["tags"] = {"items": self.data["network_tag"]}
            option["metadata"] = {
                "items": [
                    {"key": "startup-script",
                     "value":
                     self.scripts[wn_type]["startup"][machine["core"]]},
                    {"key": "shutdown-script",
                     "value":
                     self.scripts[wn_type]["shutdown"][machine["core"]]},
                ]
            }
            option["scheduling"] = {
                "preemptible": bool(self.data["preemptible"])
            }
        if "ssd" in machine:
            ssd = machine["ssd"]
            if type(ssd) is not list:
                ssd = [ssd]
            for s in ssd:
                option["disks"].append({
                    "type": "SCRATCH",
                    "boot": True,
                    "autoDelete": True,
                    "interface":  s,
                    "initializeParams": {
                        "diskType": "zones/%s/diskTypes/local-ssd"
                        % self.data["zone"]
                    }
                })
        for opt in machine:
            if opt not in ["name", "core", "mem", "disk", "image", "max",
                           "idle", "ssd"]:
                option[opt] = machine[opt]

        if wn_type is not None:
            self.wn_starting.append(instance_name)
        try:
            return self.get_gce().create_instance(instance=instance_name,
                                                  option=option,
                                                  n_wait=n_wait,
                                                  update=update)
        except HttpError as e:
            if instance_name in self.wn_starting:
                self.wn_starting.remove(instance_name)
            if e.resp.status == 409:
                self.logger.warning(e)
                return False
            raise HttpError(e.message)

    def prepare_wns(self):
        created = False
        for machine in self.data["machines"]:
            if not self.check_for_core(machine):
                continue

            if self.start_terminated(machine["core"]):
                continue

            n = 1
            while n < self.data["instance_max_num"]:
                instance_name = ("%s-%0"
                                 + str(len(str(self.data["instance_max_num"])))
                                 + "d").format() % (
                    self.prefix_core[machine["core"]], n)
                if instance_name in self.get_full_wns():
                    n += 1
                    continue

                self.new_instance(instance_name, machine, n_wait=self.n_wait,
                                  wn_type="wn")
                created = True
                break

        return created

    def check_for_test_core(self, machine):
        core = machine["core"]
        n_test_idle_jobs = self.test_idle_jobs[core] \
            if core in self.test_idle_jobs else 0
        if n_test_idle_jobs == 0:
            return False

        n_idle_jobs = self.full_idle_jobs[core] \
            if core in self.full_idle_jobs else 0
        n_primary_idle_jobs = n_idle_jobs - n_test_idle_jobs

        unclaimed = {x: y for x, y in self.wn_status.items()
                     if x.startswith(self.prefix_core[core])
                     and y == "Unclaimed"}
        test_unclaimed = {x: y for x, y in self.wn_status.items()
                          if x.startswith(self.test_prefix_core[core])
                          and y == "Unclaimed"}
        n_unclaimed = len(unclaimed)
        for wn in self.wn_starting:
            if wn.startswith(self.prefix_core[core]):
                n_unclaimed += 1
        n_unclaimed -= n_primary_idle_jobs
        if n_unclaimed < 0:
            n_unclaimed = 0

        n_unclaimed += len(test_unclaimed)
        for wn in self.wn_starting:
            if wn.startswith(self.test_prefix_core[core]):
                n_unclaimed += 1
        n_unclaimed -= n_test_idle_jobs

        if n_unclaimed >= 0:
            return False

        if self.data["max_cores"] > 0:
            if self.total_core_use + core > self.data["max_cores"]:
                return False

        return True

    def start_test_terminated(self, core):
        if self.data["reuse"] != 1:
            return False

        for instance in self.get_instances_non_terminated(update=False):
            if instance.startswith(self.prefix_core[core]) \
                    or instance.startswith(self.test_prefix_core[core]):
                self.wn_starting.append(instance)
                try:
                    self.get_gce().start_instance(instance, n_wait=self.n_wait,
                                                  update=False)
                except HttpError as e:
                    self.wn_starting.remove(instance)
                    self.logger.warning(e)
                    return False
                return True
        return True

    def prepare_test_wns(self):
        created = False
        for machine in self.data["machines"]:
            if not self.check_for_test_core(machine):
                continue

            if self.start_test_terminated(machine["core"]):
                continue

            n = 1
            while n < self.data["instance_max_num"]:
                instance_name = ("%s-%0"
                                 + str(len(str(self.data["instance_max_num"])))
                                 + "d").format() % (
                    self.test_prefix_core[machine["core"]], n)
                if instance_name in self.get_full_wns():
                    n += 1
                    continue

                self.new_instance(instance_name, machine, n_wait=self.n_wait,
                                  wn_type="test_wn")
                created = True
                break

        return created

    def prepare_wns_wrapper(self):
        self.logger.debug("prepare_wns_wrapper")
        self.full_idle_jobs, self.test_idle_jobs \
            = self.condor.condor_idle_jobs(
                exclude_owners=self.data["primary_accounts"])
        self.wn_status = self.condor.condor_wn_status()
        self.logger.debug("full_idle_jobs:" + pformat(self.full_idle_jobs))
        self.logger.debug("test_idle_jobs:" + pformat(
            self.test_idle_jobs))
        self.logger.debug("wn_status:" + pformat(self.wn_status))
        while True:
            if not self.prepare_wns():
                break
        while True:
            if not self.prepare_test_wns():
                break

    def series(self):
        self.logger.debug("series start")
        self.check_condor_status()
        self.update_config()
        self.get_instances_gce(update=True)
        self.check_required()
        self.update_wns()
        self.prepare_wns_wrapper()
        self.logger.debug("instances:\n" + pformat(list(self.instances_gce)))
        self.logger.debug("condor_wns:\n" + pformat(self.condor_wns))
        self.logger.debug("wns:\n" + pformat(self.wns))
        self.logger.debug("wn_starting:\n" + pformat(self.wn_starting))
        self.logger.debug("wn_deleting:\n" + pformat(self.wn_deleting))

    def run(self, oneshot=False):
        self.logger.info("Starting")
        while True:
            try:
                self.series()
                if oneshot:
                    break
                sleep(self.data["interval"])
            except KeyboardInterrupt:
                break
            except Exception as e:
                self.logger.error(e)
                break



# -*- coding: utf-8 -*-

"""
    Module to provide google service management
"""

import os
import sys
from .utils import expand
from .__init__ import __version__
from .__init__ import __program__
import httplib2
from google.oauth2 import service_account
from oauth2client.file import Storage
from oauth2client.tools import run_flow
from oauth2client.tools import argparser
from oauth2client.client import OAuth2WebServerFlow
from apiclient.discovery import build

__CLIENT_ID__ =\
    "154689602688-rmimpcfc5d5th2nb2rbap02sujh0ehtg.apps.googleusercontent.com"
__CLIENT_SECRET__ = "LouXW0cr1pkCoi8QtTOweld2"


def service_from_oauth(oauth_file, api_name, api_version, scope):
    oauth_file = expand(oauth_file)
    oauth_dir = os.path.dirname(oauth_file)
    if not os.path.isdir(oauth_dir):
        os.makedirs(oauth_dir)
    storage = Storage(oauth_file)
    credentials = storage.get()
    if credentials is None or credentials.invalid:
        args, unknown = argparser.parse_known_args(sys.argv)
        credentials = run_flow(
            OAuth2WebServerFlow(
                client_id=__CLIENT_ID__,
                client_secret=__CLIENT_SECRET__,
                scope=scope,
                user_agent=__program__ + '/' + __version__),
            storage, args)

    http = httplib2.Http()
    http = credentials.authorize(http)
    service = build(api_name, api_version, http=http, cache_discovery=False)
    return service


def service_from_service_account(
        service_account_file, api_name, api_version, scope):
    f = expand(service_account_file)
    credentials = service_account.Credentials.from_service_account_file(
        f, scopes=scope)
    service = build(api_name, api_version, credentials=credentials,
                    cache_discovery=False)
    return service


def get_service(service_account_file="",
                oauth_file="~/.config/gcpm/oauth",
                scope=["https://www.googleapis.com/auth/cloud-platform"],
                api_name="compute",
                api_version="v1"):
    if service_account_file == "":
        service = service_from_oauth(
            oauth_file=oauth_file, api_name=api_name, api_version=api_version,
            scope=scope)
    else:
        service = service_from_service_account(
            service_account_file=service_account_file,
            api_name=api_name, api_version=api_version, scope=scope)
    return service


def get_compute(service_account_file="",
                oauth_file="~/.config/gcpm/oauth"):
    return get_service(service_account_file=service_account_file,
                       oauth_file=oauth_file,
                       scope=[
                           "https://www.googleapis.com/auth/compute"
                       ],
                       api_name="compute",
                       api_version="v1")


def get_storage(service_account_file="",
                oauth_file="~/.config/gcpm/oauth"):
    return get_service(service_account_file=service_account_file,
                       oauth_file=oauth_file,
                       scope=[
                           "https://www.googleapis.com/auth/cloud-platform"
                       ],
                       api_name="storage",
                       api_version="v1")



# -*- coding: utf-8 -*-

"""
    Module to manage HTCondor information
"""


from .utils import proc


class Condor(object):

    def __init__(self, test=False):
        self.test = test

    def condor_q(self, opt=[]):
        return proc(["condor_q"] + opt)

    def condor_status(self, opt=[]):
        if self.test:
            return (0, "", "")
        return proc(["condor_status"] + opt)

    def condor_config_val(self, opt=[]):
        if self.test:
            return (0, "", "")
        return proc(["condor_config_val"] + opt)

    def condor_reconfig(self, opt=[]):
        if self.test:
            return (0, "", "")
        return proc(["condor_reconfig"] + opt)

    def condor_wn(self):
        if self.test:
            return ["gcp-test-wn-1core-0001"]
        wn_candidates = self.condor_status(["-autoformat", "Name"])[1].split()
        wn_candidates = [x.split(".")[0] for x in wn_candidates]
        wn_candidates2 = []
        for wn in wn_candidates:
            if "@" in wn:
                wn_candidates2.append(wn.split("@")[1])
            else:
                wn_candidates2.append(wn)
        wn_list = list(set(wn_candidates2))
        return wn_list

    def condor_wn_exist(self, wn_name):
        if self.test:
            if wn_name == "gcp-test-wn-1core-0001":
                return True
            else:
                return False

        if wn_name in self.condor_wn():
            return True
        else:
            return False

    def condor_wn_status(self):
        if self.test:
            return {"gcp-test-wn-1core-0001": "Claimed"}
        status_ret = self.condor_status(["-autoformat", "Name", "State"])[1]
        status_dict = {}
        for line in status_ret.splitlines():
            name, status = line.split()
            name = name.split(".")[0]
            if "@" in name:
                name = name.split("@")[1]
            status_dict[name] = status
        return status_dict

    def condor_idle_jobs(self, owners=[], exclude_owners=[]):
        if self.test:
            return [{1: 1}, {}]
        qinfo = self.condor_q(["-allusers", "-global", "-autoformat",
                               "JobStatus", "RequestCpus", "Owner"])[1]
        full_idle_jobs = {}
        selected_idle_jobs = {}
        if qinfo == "All queues are empty\n":
            return [full_idle_jobs, selected_idle_jobs]
        for line in qinfo.splitlines():
            status, core, owner = line.split()
            status = int(status)
            core = int(core)
            if status != 1:
                continue
            if core not in full_idle_jobs:
                full_idle_jobs[core] = 0
            full_idle_jobs[core] += 1
            if len(owners) == 0 and len(exclude_owners) == 0:
                continue
            if len(owners) > 0:
                is_owner = 0
                for o in owners:
                    if owner.startswith(o):
                        is_owner = 1
                        break
                if is_owner == 0:
                    continue
            if len(exclude_owners) > 0:
                is_owner = 1
                for o in exclude_owners:
                    if owner.startswith(o):
                        is_owner = 0
                        break
                if is_owner == 0:
                    continue
            if core not in selected_idle_jobs:
                selected_idle_jobs[core] = 0
            selected_idle_jobs[core] += 1
        return [full_idle_jobs, selected_idle_jobs]



# -*- coding: utf-8 -*-

"""
    Command line interface for core object
"""


from .core import Gcpm
import fire


class CliObject(object):
    """HTCondor pool manager for Google Cloud Platform."""

    def __init__(self, config="", test=False, oneshot=False):
        self.config = config
        self.test = test
        self.oneshot = oneshot

    def help(self):
        Gcpm.help()

    def version(self):
        Gcpm.version()

    def show_config(self):
        Gcpm(config=self.config).show_config()

    def install(self):
        """Install service related files."""
        Gcpm(config=self.config, service=True).install()

    def uninstall(self):
        """Uninstall service related files."""
        Gcpm(config=self.config, service=True).uninstall()

    def run(self):
        """Main function to run the loop."""
        Gcpm(config=self.config, test=self.test).run(oneshot=self.oneshot)

    def service(self):
        """Run the loop as service."""
        Gcpm(config=self.config, test=self.test, service=True).run()

    def set_pool_password(self, pool_password):
        """Set pool_password file in google storage."""
        Gcpm(config=self.config).get_gcs().upload_file(pool_password)


def cli():
    fire.Fire(CliObject)



# -*- coding: utf-8 -*-

"""
    Module to provide utilities
"""


def expand(path):
    import os
    return os.path.expandvars(os.path.expanduser(path))


def proc(cmd):
    import sys
    import shlex
    import subprocess
    if type(cmd) != list:
        cmd = shlex.split(cmd)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if sys.version_info.major > 2:
        stdout = stdout.decode()
        stderr = stderr.decode()
    return (p.returncode, stdout, stderr)


def make_startup_script(core, mem, disk, image, preemptible, admin,
                        head, port, domain, owner, bucket, off_timer=0,
                        wn_type=""):
    if wn_type == "test_wn":
        start = "IsPrimaryJob =!= True"
    else:
        start = "SlotID == 1"
    content = """#!/usr/bin/env bash
echo "{{\\"date\\": \\"$(date +%s)\\", \\"core\\": {core},\\"mem\\": {mem}, \
\\"disk\\": {disk}, \\"image\\": \\"{image}\\", \
\\"preemptible\\": {preemptible}}}" >/var/log/nodeinfo.log

sed -i"" 's/FIXME_ADMIN/{admin}/' /etc/condor/config.d/00_config_local.config

sed -i"" 's/FIXME_HOST/{head}/' /etc/condor/config.d/10_security.config
sed -i"" 's/FIXME_PORT/{port}/' /etc/condor/config.d/10_security.config
sed -i"" 's/FIXME_DOMAIN/{domain}/' /etc/condor/config.d/10_security.config
sed -i"" "s/FIXME_PRIVATE_DOMAIN/$(hostname -d)/" \
/etc/condor/config.d/10_security.config

sed -i"" 's/FIXME_OWNER/{owner}/' /etc/condor/config.d/20_workernode.config
sed -i"" 's/FIXME_CORE/{core}/' /etc/condor/config.d/20_workernode.config
sed -i"" 's/FIXME_MEM/{mem}/' /etc/condor/config.d/20_workernode.config

sed -i"" 's/FIXME_START/{start}/' /etc/condor/config.d/20_workernode.config

gsutil cp "gs://{bucket}/pool_password" /etc/condor/
chmod 600 /etc/condor/pool_password
systemctl enable condor
systemctl start condor
while :;do
  condor_reconfig
  status="$(condor_status | grep "${{HOSTNAME}}")"
  if [ -n "$status" ];then
    break
  fi
  sleep 10
done
date >> /root/condor_started""".format(core=core, mem=mem, disk=disk,
                                       image=image, preemptible=preemptible,
                                       admin=admin, head=head, port=port,
                                       domain=domain, owner=owner,
                                       bucket=bucket, start=start)

    if off_timer != 0:
        content += """
sleep {off_timer}
condor_off -peaceful -startd
date >> /root/condor_off""".format(off_timer=off_timer)
    return content


def make_shutdown_script():
    content = """#!/usr/bin/env bash
preempted=$(\
curl "http://metadata.google.internal/computeMetadata/v1/instance/preempted" \
-H "Metadata-Flavor: Google")
echo "{{\\"date\\": \\"$(date +%s)\\", \\"preempted\\": ${{preempted}}}}" \
>>/var/log/shutdown.log""".format()
    return content



# -*- coding: utf-8 -*-

"""
    Module to mange Google Cloud Storage.
"""


import os
import logging
import googleapiclient
from .service import get_storage
from .utils import expand


class Gcs(object):

    def __init__(self, oauth_file="", service_account_file="", project="",
                 storageClass="", location="", bucket=""):
        self.oauth_file = oauth_file
        self.service_account_file = service_account_file
        self.project = project
        self.storageClass = storageClass
        self.location = location
        self.bucket = bucket
        self.storage_service = None
        self.logger = logging.getLogger(__name__)

    def storage(self):
        if self.storage_service is None:
            self.storage_service = get_storage(
                service_account_file=self.service_account_file,
                oauth_file=self.oauth_file,
            )
        return self.storage_service

    def is_bucket(self):
        return True if self.bucket in self.get_buckets() else False

    def get_buckets(self):
        bucket_list = [x["name"] for x in self.storage().buckets().list(
            project=self.project).execute()["items"]]
        return bucket_list

    def delete_bucket(self):
        if not self.is_bucket():
            return
        self.storage().buckets().delete(bucket=self.bucket).execute()

    def create_bucket(self):
        if self.is_bucket():
            return
        body = {"name": self.bucket,
                "sotrageClass": self.storageClass,
                "location": self.location,
                }
        return self.storage().buckets().insert(project=self.project,
                                               body=body).execute()

    def upload_file(self, file_name):
        self.create_bucket()
        with open(expand(file_name), 'rb') as f:
            response = self.storage().objects().insert(
                bucket=self.bucket,
                media_body=googleapiclient.http.MediaIoBaseUpload(
                    f, 'application/octet-stream'),
                name=os.path.basename(file_name)).execute()
        return response

    def delete_file(self, file_name):
        return self.storage().objects().delete(
            bucket=self.bucket, object=file_name).execute()



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['gcpm']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.1.3,<0.2.0',
 'google-api-python-client>=1.7,<2.0',
 'google-auth-httplib2>=0.0.3,<0.0.4',
 'google-auth>=1.6,<2.0',
 'oauth2client>=4.1,<5.0',
 'ruamel.yaml>=0.15.83,<0.16.0']

entry_points = \
{'console_scripts': ['gcpm = gcpm:main']}

setup_kwargs = {
    'name': 'gcpm',
    'version': '0.2.2',
    'description': 'Google Cloud Platform Condor Pool Manager',
    'long_description': '# Google Cloud Platform Condor Pool Manager (GCPM)\n\n[![Build Status](https://travis-ci.org/mickaneda/gcpm.svg?branch=master)](https://travis-ci.org/mickaneda/gcpm) ([Coverage report](https://mickaneda.github.io/gcpm/))\n\nHTCondor pool manager for Google Cloud Platform.\n\n## Installation\n\n### Package installation\n\nGCPM can be installed by `pip`:\n\n    $ pip install gcpm\n\n### Service file installation\n\nTo install as service, do:\n\n    $ gcpm install\n\n:warning: Service installation is valid only for the system managed by **Systemd**.\n\nIf **logrotate** is installed, logrotation definition for **/var/log/gcpm.log** is also installed.\n\n## Configuration file\n\n### Configuration file path\n\nThe default configuration file is **~/.config/gcpm/gcpm.yml**.\n\nFor service, the configuration file is **/etc/gcpm.yml**.\n\nTo change the configuration file, use `--config` option:\n\n    $ gcpm run --config /path/to/my/gcpm.yml\n\n### Configuration file content\n\nA configuration file is YAML format.\n\nName|Description|Default Value|Mandatory|\n:---|:----------|:------------|:--------|\nconfig_dir   | Directory for some gcpm related files.|**~/.config/gcpm/** (user)<br>**/var/cache/gcpm** (service)|No\noatuh_file   | Path to OAuth information file for GCE/GCS usage.|**<config_dir>/oauth**|No\nservice_account_file | Service account JSON file for GCE/GCS usage.<br>If not specified, OAuth connection is tried.|-|No\nproject      | Google Cloud Platform Project Name.|-|Yes\nzone         | Zone for Google Compute Engine.|-|Yes\nmachines     | Array of machine settings.<br>Each setting is array of [core, mem, disk, idle, image] (see below).|[]|Yes\nmachines:core     | Number of core of the machine type.|-|Yes\nmachines:mem      | Memory (MB) of the machine type.|-|Yes\nmachines:disk     | Disk size (GB) of the machine type.|-|Yes\nmachines:max      | Limit of the number of instances for the machine type.|-|Yes\nmachines:idle     | Number of idle machines for the machine type.|-|Yes\nmachines:image    | Image of the machine type.|-|Yes\nmachines:&lt;others&gt; | Other any options can be defined for creating instance.|-|No\nmax_cores    | Limit of the total number of cores of all instances.<br>If it is set 0, no limit is applied.|0|No\nstatic_wns   | Array of instance names of static worker nodes, which are added as condor worker nodes.|[]|No\nrequired_machines          | Array of machines which should be running other than worker nodes.|[]|No\nrequired_machines:name     | Number of core of the machine type.|-|Yes\nrequired_machines:mem      | Memory (MB) of the machine type.|-|Yes\nrequired_machines:disk     | Disk size (GB) of the machine type.|-|Yes\nrequired_machines:image    | Image of the machine type.|-|Yes\nrequired_machines:&lt;others&gt; | Other any options can be defined for creating instance.|-|No\nprimary_accounts |User accounts which jobs must run normal worker nodes. See below about primary accounts.|[]|No\nprefix       | Prefix of machine names.|**gcp-wn**|No\npreemptible  | 1 for preemptible machines, 0 for not.|0|No\noff_timer    | Second to send condor_off after starting.|0|No\nnetwork_tag  | Array of GCP network tag.|[]|No\nreuse        | 1 to reused terminated instance. Otherwise delete and re-created instances.|0|No\ninterval     | Second of interval for each loop.|10|No\nhead_info    | If **head** is empty, head node information is automatically taken for each option:<br>hostname: Hostname<br>ip: IP address<br>gcp: Hostname|**gcp**|No\nhead         | Head node Hostname/IP address.|""|No\nport         | HTCondor port.|9618|No\ndomain       | Domain of the head node.<br>Set empty to take it from hostnaem.|""|No\nadmin        | HTCondor admin email address.|""|Yes\nowner        | HTCondor owner name.|""|Yes\nwait_cmd     | 1 to wait GCE commands result (create/start/stop/delete...).|0|No\nbucket       | Bucket name for pool_password file.|""|Yes\nstorageClass | Storage class name of the bucket.|"REGIONAL"|No\nlocation     | Storage location for the bucket.<br>If empty, it is decided from the **zone**.|""|No\nlog_file     | Log file path. Empty to put it in stdout.|""|No\nlog_level    | Log level. (**debug**, **info**, **warning**, **error**, **critical**)|**info**|No\n\n\nNote:\n\n* Primary accounts\n\nIf primary accounts are set, jobs of **non-primary** accounts can run on test worker nodes.\n\nIf there are already max number of 1 core worker nodes\nand idle jobs of non-primary accounts are there,\ntest worker node named **&lt;prefix&gt-test-1core-XXXX** will be launched\nand only non-primary account jobs can run on it.\n\nThis able to run such a test job w/o waiting for finishing any normal jobs.\n\nSuch test worker nodes can be launched until total cores are smaller than `max_core`.\n\nTo use this function effectively, set total of `max` of each core to less than `max_core`.\n\ne.g.)\n\n```yml\n---\nmachines:\n  core: 1\n  max: 10\nmachines:\n  core: 8\n  max:  2\nmax_core: 20\nprimary_accounts:\n  - condor_primary\n```\n\nIn this case, normal jobs can launch 10 1 core machines and 2 8 core machines,\nthen 16 cores are used.\n\nEven if there are a log of idle **condor_primary**\'s jobs,\n1 core test jobs by other accounts can run: 4 jobs at most.\n\n',
    'author': 'Michiru Kaneda',
    'author_email': 'Michiru.Kaneda@gmail.com',
    'url': 'https://github.com/mickaneda/gcpm',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

"""
    Module to mange Google Cloud Storage.
"""


import os
from .utils import expand, proc


__SERVICE_FILE__ = "/usr/lib/systemd/system/gcpm.service"
__LOGROTATE_FILE__ = "/etc/logrotate.d/gcpm.conf"


def make_file(filename, content="", mkdir=True):
    filename = expand(filename)
    directory = os.path.dirname(filename)
    if not os.path.isdir(directory):
        if mkdir:
            os.makedirs(directory)
        else:
            return False
    with open(filename, mode='w') as f:
            f.write(content)
    return True


def rm_file(filename):
    if not os.path.isfile(filename):
        return
    os.remove(filename)


def make_service(filename=__SERVICE_FILE__, mkdir=True):
    content = """[Unit]
Description = HTCondor pool manager for Google Cloud Platform

[Service]
Environment = "PATH={path}"
ExecStart = /usr/bin/gcpm service
ExecStop = /usr/bin/kill -p $MAINPID
Restart = always
StandardOutput = journal
StandardError = journal
SyslogIdentifier = gcpm

[Install]
WantedBy = multi-user.target""".format(path=os.environ["PATH"])
    make_file(filename, content, mkdir)
    proc(["systemctl", "daemon-reload"])


def rm_service(filename=__SERVICE_FILE__):
    rm_file(filename)


def make_logrotate(filename=__LOGROTATE_FILE__, mkdir=True):
    content = """/var/log/gcpm.log {{
  missingok
  rotate 10
  dateext
  delaycompress
  daily
  minsize 100M
  postrotate
      systemctl restart gcpm
  endscript
}}""".format()
    make_file(filename, content, mkdir)


def rm_logrotate(filename=__LOGROTATE_FILE__):
    rm_file(filename)
